
from binascii import hexlify
from collections import namedtuple
from inspect import stack, getmodule
from json import loads


class WarnDict(dict):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except KeyError:
            msg = self.__msg % (item,)
            self.__log.debug(msg)
            raise KeyError(msg)


class WarnList(list):

    def __init__(self, log, msg):
        self.__log = log
        self.__msg = msg
        super().__init__()

    def __getitem__(self, item):
        try:
            return super().__getitem__(item)
        except IndexError:
            msg = self.__msg % item
            self.__log.debug(msg)
            raise IndexError(msg)


def tohex(data):
    return hexlify(data).decode('ascii')


def assert_attr(instance, *attrs):
    for attr in attrs:
        if getattr(instance, attr) is None:
            raise Exception('No %s defined' % attr)


def kargs_to_attr(**kargs):
    return dict_to_attr(kargs)


def dict_to_attr(kargs):
    return namedtuple('Attr', kargs.keys(), rename=True)(*kargs.values())


class MutableAttr(dict):

    def __init__(self, *args, none=False, **kargs):
        self.__none = none
        super().__init__(*args, **kargs)

    def __getattr__(self, name):
        try:
            return self[name]
        except KeyError:
            if self.__none:
                return None
            else:
                raise AttributeError(name)

    def __setattr__(self, name, value):
        if name.startswith('_'):
            super().__setattr__(name, value)
        else:
            self[name] = value

    def _to_dict(self):
        return self.__dict__


class MissingReference(Exception): pass


def reftuple(name, *args, **kargs):
    '''
    Like a namedtuple, but expands $ strings using a database session and date
    (# is similar, but also does JSON parsing).
    '''

    from ..squeal import StatisticJournal, StatisticName

    class klass(namedtuple(name, *args, **kargs)):

        def expand(self, log, s, time, owner=None, constraint=None):
            instance = self
            for name in self._fields:
                value = getattr(instance, name)
                if isinstance(value, str) and len(value) > 0 and value[0] in '$#':
                    log.info(f'Expanding {value} for {name}')
                    replacement = self._lookup(log, s, time, value[1:], value[0] == '#',
                                               default_value=self._fields_defaults.get(name, None),
                                               default_owner=owner, default_constraint=constraint)
                    instance = instance._replace(**{name: replacement})
            return instance

        def _lookup(self, log, s, time, name, json,
                    default_value=None, default_owner=None, default_constraint=None):
            if isinstance(name, str) and len(name) > 0 and name[0] in '$#':
                name = self._lookup(log, s, time, name[1:], name[0] == '#',
                                    default_owner=default_owner, default_constraint=default_constraint)
            owner, name, constraint = StatisticName.parse(name, default_owner=default_owner,
                                                          default_constraint=default_constraint)
            value = StatisticJournal.before(s, time, name, owner, constraint)
            if value is None:
                if default_value:
                    log.warning(f'No value found for {owner}:{name}:{constraint} (default {default_value})')
                    value = default_value
                else:
                    raise MissingReference(f'No value found for {owner}:{name}:{constraint} (and no default)')
            else:
                value = value.value
                if json:
                    log.debug(f'Unpacking JSON "{value}"')
                    value = loads(value)
                log.info(f'{name} -> {value}')
            return value

    klass.__name__ = name
    caller = stack()[1]
    klass.__module__ = getmodule(None, caller.filename).__name__
    return klass


class MaxDict(dict):

    def __init__(self, kv):
        super().__init__()
        for key, value in kv:
            if key in self:
                self[key] = max(value, self[key])
            else:
                self[key] = value


def nearest_index(df, name, value):
    exactmatch = df.loc[df[name] == value]
    if not exactmatch.empty:
        return exactmatch.index[0]
    else:
        lower = df.loc[df[name] < value].index.dropna()
        upper = df.loc[df[name] > value].index.dropna()
        if lower.empty:
            return upper.min()
        elif upper.empty:
            return lower.max()
        else:
            if abs(value - df.loc[lower.max()][name]) < abs(value - df.loc[upper.min()][name]):
                return lower.max()
            else:
                return upper.min()


def get_index_loc(df, value):
    loc = df.index.get_loc(value)
    try:
        return loc.start  # if slice, take first
    except AttributeError:
        return loc  # otherwise, simple value




import datetime as dt
from collections import deque, namedtuple
from logging import getLogger

from sqlalchemy import select, and_
from sqlalchemy.sql.functions import coalesce

from ..squeal.tables.statistic import StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat
from ..squeal.utils import tables

log = getLogger(__name__)


def make_waypoint(names, extra=None):
    names = list(names)
    if extra:
        names += [extra]
    names = ['time'] + names
    defaults = [None] * len(names)
    return namedtuple('Waypoint', names, defaults=defaults)


class WaypointReader:

    def __init__(self, with_timespan=True):
        self._with_timespan = with_timespan

    def read(self, s, ajournal, names, owner=None, start=None, finish=None):

        t = tables(StatisticName, StatisticJournal, StatisticJournalInteger, StatisticJournalFloat)

        id_map = self._id_map(s, ajournal, names, owner=owner)
        ids = list(id_map.keys())

        Waypoint = make_waypoint(names.values(), extra='timespan' if self._with_timespan else None)

        for timespan in ajournal.timespans:
            log.debug('%s' % timespan)
            # log.debug(f'{timespan.start.timestamp()} - {timespan.finish.timestamp()}')
            waypoint = None
            stmt = select([t.StatisticName.c.id,
                           t.StatisticJournal.c.time,
                           coalesce(t.StatisticJournalFloat.c.value,
                                    t.StatisticJournalInteger.c.value)]) \
                .select_from(t.StatisticJournal.
                             join(t.StatisticName).
                             outerjoin(t.StatisticJournalFloat).
                             outerjoin(t.StatisticJournalInteger)) \
                .where(and_(t.StatisticJournal.c.source_id == ajournal.id,
                            t.StatisticName.c.id.in_(ids),
                            t.StatisticJournal.c.time >= timespan.start,
                            t.StatisticJournal.c.time <= timespan.finish)) \
                .order_by(t.StatisticJournal.c.time)
            if start:
                stmt = stmt.where(t.StatisticJournal.c.time >= start)
            if finish:
                stmt = stmt.where(t.StatisticJournal.c.time <= finish)
            # log.debug(stmt)
            for id, time, value in s.connection().execute(stmt):
                if waypoint and waypoint.time != time:
                    # log.debug(waypoint)
                    yield waypoint
                    waypoint = None
                if not waypoint:
                    waypoint = Waypoint(time=time)
                    if self._with_timespan:
                        waypoint = waypoint._replace(timespan=timespan)
                waypoint = waypoint._replace(**{id_map[id]: value})
        log.debug('Waypoints generated')

    def _id_map(self, s, ajournal, names, owner):
        # need to convert from statistic_name_id to attribute name
        return dict((self._id(s, ajournal, key, owner), value) for key, value in names.items())

    def _id(self, s, ajournal, name, owner=None):
        from ..data.frame import _add_constraint
        q = s.query(StatisticName.id). \
            filter(StatisticName.name == name, StatisticName.constraint == ajournal.activity_group)
        if owner:
            q = _add_constraint(q, StatisticName.owner, owner.split(','), name)
        return q.scalar()


class Chunk:
    '''
    A collection of data points in time order, associated with a single timespan.

    In most of uses the contents are slowly incremented over time (and values popped off the front)
    as various statistics are calculated.
    '''

    def __init__(self, waypoint):
        self.__timespan = waypoint.timespan
        self.__waypoints = deque([waypoint])

    def append(self, waypoint):
        self.__waypoints.append(waypoint)

    def popleft(self):
        return self.__waypoints.popleft()

    def __diff(self, index, attr, zero=0):
        if len(self.__waypoints) > 1:
            return attr(self.__waypoints[index]) - attr(self.__waypoints[0])
        else:
            return zero

    def distance(self):
        return self.__diff(-1, lambda w: w.distance)

    def distance_delta(self):
        return self.__diff(1, lambda w: w.distance)

    def time(self):
        return self.__diff(-1, lambda w: w.time, dt.timedelta(0)).total_seconds()

    def time_delta(self):
        return self.__diff(1, lambda w: w.time, dt.timedelta(0)).total_seconds()

    def values(self, name):
        index = self.__waypoints[0]._fields.index(name)
        return (waypoint[index] for waypoint in self.__waypoints if waypoint[index] is not None)

    def __len__(self):
        return len(self.__waypoints)

    def __getitem__(self, item):
        return self.__waypoints[item]

    def __bool__(self):
        return self.distance() > 0


class Chunks:
    '''
    This returns a sequence of lists of chunks, with progressively more waypoints.  The chunks are
    the *same* list, but with more points added each time (ie mutated / extended),

    The caller should read chunks until they contain sufficient data (for whatever is being calculated)
    and then discard and call to maintain the same data length.
    '''

    def __init__(self, waypoints):
        self._waypoints = waypoints

    def chunks(self):
        chunks, chunk_index = deque(), {}
        for waypoint in self._waypoints:
            timespan = waypoint.timespan
            if timespan in chunk_index:
                chunk_index[timespan].append(waypoint)
            else:
                chunk = Chunk(waypoint)
                chunk_index[timespan] = chunk
                chunks.append(chunk)
            yield chunks

    def complete(self):
        '''
        The final chunks, containing all waypoints.
        '''
        last = None
        for chunks in self.chunks():
            last = chunks
        return last

    @staticmethod
    def drop_first(chunks):
        chunks[0].popleft()
        while chunks and not chunks[0]:
            chunks.popleft()


def filter_none(names, waypoints):
    names = list(names)
    return [w for w in waypoints if all(n in w._fields and getattr(w, n) is not None for n in names)]




import datetime as dt
from calendar import monthrange
from logging import getLogger

from colorama import init, Fore, Style

from ch2.lib.schedule import Schedule
from ..lib.date import to_date, add_date, MONTH
from ..commands.args import SCHEDULE, MONTHS, START

log = getLogger(__name__)
INDENT = '   '


def test_schedule(args, db):
    '''
## test-schedule

    > ch2 test-schedule SCHEDULE

Print a calendar showing how the given schedule is interpreted.

### Example

    > ch2 test-schedule 2w[1mon,2sun]

(Try it and see)
    '''
    schedule, start, months = Schedule(args[SCHEDULE]), to_date(args[START], none=True), args[MONTHS]
    if not start:
        start = dt.date.today()
    start = dt.date(start.year, start.month, 1)
    if not months:
        months = 3
    print_calendar(schedule, start, months)


def print_calendar(schedule, start, months):
    init()
    frame, next_frame = 0, schedule.next_frame(start)
    for _ in range(months):
        print()
        print(INDENT, end='')
        title = start.strftime('%B %Y')
        print(' ' * (11 - len(title) // 2) + title)
        print(INDENT, end='')
        for day in ('Mo', 'Tu', 'We', 'Th', 'Fr', 'Sa', 'Su'):
            print(' %s' % day, end='')
        print()
        for week in month_days(start):
            print(INDENT, end='')
            for day in week:
                if not day:
                    print('   ', end='')
                else:
                    date = dt.date(start.year, start.month, day)
                    if date >= next_frame:
                        frame += 1
                        next_frame = schedule.next_frame(date)
                    if schedule.at_location(date):
                        colour = Fore.RED if frame % 2 else Fore.GREEN
                        print(colour + ' %2d' % day + Style.RESET_ALL, end='')
                    else:
                        print(' %2d' % day, end='')
            print()
        start = add_date(start, (1, MONTH))
    print()


def month_days(date):
    dow, n = monthrange(date.year, date.month)
    start = -dow
    while start < n:
        week = []
        if start < 0:
            week += [None] * -start
        week += list(range(max(0, start) + 1, min(start+7, n) + 1))
        week += [None] * (7-len(week))
        yield week
        start += 7




import re
from json import loads
from logging import getLogger

from . import ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator
from ..names import ELEVATION, DISTANCE, M, POWER_ESTIMATE, HEART_RATE, ACTIVE_DISTANCE, MSR, SUM, CNT, MAX, summaries, \
    ACTIVE_SPEED, ACTIVE_TIME, AVG, S, KMH, MIN_KM_TIME_ANY, MIN, MED_KM_TIME_ANY, PERCENT_IN_Z_ANY, PC, \
    TIME_IN_Z_ANY, MAX_MED_HR_M_ANY, W, BPM, MAX_MEAN_EP_M_ANY, CLIMB_ELEVATION, CLIMB_DISTANCE, CLIMB_TIME, \
    CLIMB_GRADIENT, TOTAL_CLIMB, HR_ZONE, TIME, like
from ...data.activity import active_stats, times_for_distance, hrz_stats, max_med_stats, max_mean_stats
from ...data.climb import find_climbs, Climb
from ...data.frame import activity_statistics, present
from ...squeal import StatisticJournalFloat, Constant

log = getLogger(__name__)


class ActivityCalculator(ActivityJournalCalculatorMixin, DataFrameCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, cost_calc=20, cost_write=1, climb=None, **kargs):
        self.climb_ref = climb
        super().__init__(*args, cost_calc=cost_calc, cost_write=cost_write, **kargs)

    def _read_dataframe(self, s, ajournal):
        try:
            return activity_statistics(s, DISTANCE, ELEVATION, HEART_RATE, HR_ZONE, POWER_ESTIMATE,
                                       activity_journal=ajournal, with_timespan=True)
        except Exception as e:
            log.warning(f'Failed to generate statistics for activity: {e}')
            raise

    def _calculate_stats(self, s, ajournal, df):
        stats, climbs = {}, None
        stats.update(active_stats(df))
        stats.update(times_for_distance(df))
        stats.update(hrz_stats(df))
        stats.update(max_med_stats(df))
        stats.update(max_mean_stats(df))
        if present(df, ELEVATION):
            params = Climb(**loads(Constant.get(s, self.climb_ref).at(s).value))
            climbs = list(find_climbs(df, params=params))
        return df, stats, climbs

    def _copy_results(self, s, ajournal, loader, data):
        df, stats, climbs = data
        self.__copy(ajournal, loader, stats, ACTIVE_DISTANCE, M, summaries(MAX, CNT, SUM, MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, ACTIVE_TIME, S, summaries(MAX, SUM, MSR), ajournal.start)
        self.__copy(ajournal, loader, stats, ACTIVE_SPEED, KMH, summaries(MAX, AVG, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, MIN_KM_TIME_ANY, S, summaries(MIN, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, MED_KM_TIME_ANY, S, summaries(MIN, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, PERCENT_IN_Z_ANY, PC, None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, TIME_IN_Z_ANY, S, None, ajournal.start)
        self.__copy_all(ajournal, loader, stats, MAX_MED_HR_M_ANY, BPM, summaries(MAX, MSR), ajournal.start)
        self.__copy_all(ajournal, loader, stats, MAX_MEAN_EP_M_ANY, W, summaries(MAX, MSR), ajournal.start)
        if climbs:
            loader.add(TOTAL_CLIMB, M, summaries(MAX, SUM, MSR), ajournal.activity_group, ajournal,
                       sum(climb[CLIMB_ELEVATION] for climb in climbs), ajournal.start, StatisticJournalFloat)
            for climb in sorted(climbs, key=lambda climb: climb[TIME]):
                self.__copy(ajournal, loader, climb, CLIMB_ELEVATION, M, summaries(MAX, SUM, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_DISTANCE, M, summaries(MAX, SUM, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_TIME, S, summaries(MAX, SUM, MSR), climb[TIME])
                self.__copy(ajournal, loader, climb, CLIMB_GRADIENT, PC, summaries(MAX, SUM, MSR), climb[TIME])

    def __copy_all(self, ajournal, loader, stats, pattern, units, summary, time):
        for name in like(pattern, stats):
            self.__copy(ajournal, loader, stats, name, units, summary, time)

    def __copy(self, ajournal, loader, stats, name, units, summary, time):
        loader.add(name, units, summary, ajournal.activity_group, ajournal, stats[name], time, StatisticJournalFloat)




from urwid import LineBox, Columns

from .focus import FocusWrap
from .widgets import ColText


class Border(LineBox):
    """
    A blank border, one character in size.
    """

    def __init__(self, contents):
        super().__init__(contents, tlcorner=' ', tline=' ', lline=' ', trcorner=' ',
                         blcorner=' ', rline=' ', bline=' ', brcorner=' ')
        try:
            self.contents = self._w.contents
        except AttributeError:
            pass


class Indent(FocusWrap):
    """
    An indent on the left.
    """

    def __init__(self, contents, width=1):
        super().__init__(Columns([ColText(' ' * width), contents]))
        try:
            self.contents = self._w.contents
        except AttributeError:
            pass




import datetime as dt
from logging import getLogger

from .format.tokens import FileHeader, token_factory, Checksum, State
from .profile.profile import read_profile
from ..commands.args import ADD_HEADER, mm, HEADER_SIZE, PROFILE_VERSION, PROTOCOL_VERSION, MIN_SYNC_CNT, \
    MAX_RECORD_LEN, MAX_DROP_CNT, MAX_BACK_CNT, MAX_FWD_LEN, MAX_DELTA_T
from ..lib.date import format_time, format_seconds

log = getLogger(__name__)


def fix(data, warn=False,
        add_header=False, drop=False, slices=None, start=None, fix_header=False, fix_checksum=False, force=True,
        validate=True, header_size=None, protocol_version=None, profile_version=None, min_sync_cnt=3,
        max_record_len=None, max_drop_cnt=1, max_back_cnt=3, max_fwd_len=200, max_delta_t=None, profile_path=None):

    slices = parse_slices(slices)
    types, messages = read_profile(log, warn=warn, profile_path=profile_path)
    data = bytearray(data)

    log_data('Initial', data)

    if add_header:
        data = prepend_header(data, header_size, protocol_version, profile_version)

    if drop:
        slices = drop_data(State(types, messages, max_delta_t=max_delta_t), data, warn=warn, force=force,
                           min_sync_cnt=min_sync_cnt, max_record_len=max_record_len, max_drop_cnt=max_drop_cnt,
                           max_back_cnt=max_back_cnt, max_fwd_len=max_fwd_len)

    if slices:
        data = apply_slices(data, slices)

    if start:
        data = set_start(data, types, messages, start)

    if fix_header or fix_checksum:
        data = header_and_checksums(data, State(types, messages, max_delta_t=max_delta_t),
                                    fix_header=fix_header, fix_checksum=fix_checksum,
                                    header_size=header_size, protocol_version=protocol_version,
                                    profile_version=profile_version)

    if validate:
        validate_data(data, State(types, messages, max_delta_t=max_delta_t), warn=warn, force=force)

    log_data('Final', data)

    return data


def parse_slices(slices):
    if not slices:
        return None
    parsed = [parse_slice(slice) for slice in slices.split(',')]
    if parsed and parsed[-1].stop is None:
        parsed[-1] = slice(parsed[-1].start, -2)
    log.debug('Parsed "%s" as "%s"' % (slices, format_slices(parsed)))
    return parsed


def parse_slice(s):
    start, stop = (parse_offset(offset) for offset in s.split(':'))
    return slice(start, stop)


def parse_offset(offset):
    if offset:
        return int(offset)
    else:
        return None


def format_slices(slices):
    return ','.join(format_slice(slice) for slice in slices)


def format_slice(slice):
    return '%s:%s' % (format_offset(slice.start), format_offset(slice.stop))


def format_offset(offset):
    return '' if offset in (0, None) else str(offset)


def log_data(title, data):
    log.info('%s Data ----------' % title)
    log.info('Length: %d bytes' % len(data))
    try:
        header = FileHeader(data)
        log.info('Header size: %d' % header.header_size)
        log.info('Protocol version: %d' % header.protocol_version)
        log.info('Profile version: %d' % header.profile_version)
    except Exception as e:
        log.info('Could not parse header: %s' % e)
    try:
        checksum = Checksum(data[-2:])
        log.info('Checksum: %d (0x%04x)' % (checksum.checksum, checksum.checksum))
    except Exception as e:
        log.info('Could not parse checksum: %s' % e)


def prepend_header(data, header_size, protocol_version, profile_version):

    log.info('Add Header ----------')

    header_size = set_default(HEADER_SIZE, header_size, 14)
    protocol_version = set_default(PROTOCOL_VERSION, protocol_version, 0x10)  # from FR35
    profile_version = set_default(PROFILE_VERSION, profile_version, 0x07de)  # from FR35

    data = bytearray([header_size] + [0] * (header_size - 1)) + data
    header = FileHeader(data)
    header.repair(data, log,
                  header_size=header_size, protocol_version=protocol_version, profile_version=profile_version)
    data[:len(header)] = header.data
    return data


def log_param(name, value):
    log.info('%s %s' % (mm(name), value))


def set_default(name, value, deflt):
    if value is None:
        value = deflt
    log_param(name, value)
    return value


def header_and_checksums(data, initial_state, fix_header=False, fix_checksum=False,
                         header_size=None, protocol_version=None, profile_version=None):
    log.info('Header and Checksums ----------')
    log_param(HEADER_SIZE, header_size)
    log_param(PROTOCOL_VERSION, protocol_version)
    log_param(PROFILE_VERSION, profile_version)
    if fix_header:
        data = process_header(data, header_size=header_size, protocol_version=protocol_version,
                              profile_version=profile_version)
    if fix_checksum:
        data = process_checksum(data, initial_state.copy())
    if fix_header and fix_checksum:
        data = process_header(data)  # if length changed with checksum
        data = process_checksum(data, initial_state.copy())  # if length changed
    return data


def process_header(data, header_size=None, protocol_version=None, profile_version=None):
    try:
        with memoryview(data) as view:
            header = FileHeader(view)
            old_size = header.header_size
            # if these are undefined, use existing values (might be used if size changes)
            if protocol_version is None:
                protocol_version = header.protocol_version
            if profile_version is None:
                profile_version = header.profile_version
            prev_len = len(header)
            header.repair(data, log, header_size=header_size, protocol_version=protocol_version, profile_version=profile_version)
            new_data = header.data
        # need to mutate if length changed (in which case header.data is no longer a memoryview)
        if header.header_size != old_size:
            data[:prev_len] = new_data
        return data
    except Exception as e:
        log.error(e)
        raise Exception('Error fixing header - maybe try %s' % mm(ADD_HEADER))


def process_checksum(data, state):
    offset = 0
    try:
        # don't use memoryview here as it gets spread into state and token
        offset = len(FileHeader(data))
        while len(data) - offset > 2:
            token = token_factory(data[offset:], state)
            offset += len(token)
        if len(data) - offset < 2:
            n = offset + 2 - len(data)
            log.warning('Adding %d byte(s) for checksum' % n)
            data += bytearray([0] * n)
        with memoryview(data) as view:
            checksum = Checksum(view[offset:])
            checksum.repair(view, log)
        return data
    except Exception as e:
        log.error(e)
        raise Exception('Error fixing checksum at offset %d' % offset)


def apply_slices(data, slices):

    log.info('Slice ----------')
    log.info('Slices: %s' % format_slices(slices))

    result = bytearray()
    for slice in slices:
        result += data[slice]
    log.info('Have %d bytes after slicing' % len(result))
    dropped = len(data) - len(result)
    if dropped:
        log.warning('Slicing decreased length by %d bytes' % dropped)

    return result


class StartState(State):

    def __init__(self, types, messages, start, max_delta_t=None):
        super().__init__(types, messages, max_delta_t=max_delta_t)
        self.__start = start
        self.__delta = None

    @property
    def timestamp(self):
        return self._timestamp

    @timestamp.setter
    def timestamp(self, timestamp):
        if self.__delta is None:
            self.__delta = self.__start - timestamp
            log.warning('Shifting timestamps by %s' % format_seconds(self.__delta.total_seconds()))
        self._timestamp = self._validate_timestamp(timestamp + self.__delta)


def set_start(data, types, messages, start):

    log.info('Start ----------')
    log.info('Start: %s' % format_time(start))

    state = StartState(types, messages, start)
    with memoryview(data) as view:
        offset = len(FileHeader(view))
        while len(view) - offset > 2:
            token = token_factory(view[offset:], state)
            token.repair_timestamp(state)
            offset += len(token)

    return data


def validate_data(data, state, warn=False, force=True):

    log.info('Validation ----------')
    log_param(MAX_DELTA_T, state.max_delta_t)
    if state.max_delta_t is None:
        log.warning('Time-reversal is allowed unless %s is set' % MAX_DELTA_T)

    first_t = True
    try:
        file_header = FileHeader(data)
        file_header.validate(data, log)
        offset = len(file_header)
        while len(data) - offset > 2:
            token = token_factory(data[offset:], state)
            if first_t and state.timestamp:
                log.info('First timestamp: %s' % state.timestamp)
                first_t = False
            record = token.parse_token(warn=warn)
            if force:
                record.force()
            offset += len(token)
        log.info('Last timestamp:  %s' % state.timestamp)
        if state.timestamp > dt.datetime.now(tz=dt.timezone.utc):
            log.warning('Timestamp in future')
        checksum = Checksum(data[offset:])
        checksum.validate(data, log)
        log.info('OK')
    except Exception as e:
        log.error(e)
        log.info('Validation failed')
        raise


def drop_data(initial_state, data, warn=False, force=True,
              min_sync_cnt=3, max_record_len=None, max_drop_cnt=1, max_back_cnt=3, max_fwd_len=200):

    log.info('Drop Data ----------')
    for (name, value) in [(MIN_SYNC_CNT, min_sync_cnt), (MAX_RECORD_LEN, max_record_len),
                          (MAX_DROP_CNT, max_drop_cnt), (MAX_BACK_CNT, max_back_cnt),
                          (MAX_FWD_LEN, max_fwd_len), (MAX_DELTA_T, initial_state.max_delta_t)]:
        log_param(name, value)

    # use memoryview for efficient slicing (although it doesn't seem to help much)
    with memoryview(data) as view:
        slices = advance(initial_state, view,
                         drop_count=0, initial_offset=0, warn=warn, force=force,
                         min_sync_cnt=min_sync_cnt, max_record_len=max_record_len, max_drop_cnt=max_drop_cnt,
                         max_back_cnt=max_back_cnt, max_fwd_len=max_fwd_len)
    log.info('Found slices %s' % format_slices(slices))
    return slices


def offset_tokens(state, data, offset=0, warn=False, force=True):
    '''
    this yields the offsets *after* the tokens (unlike tokens() in the read module).
    '''
    try:
        if not offset:
            file_header = FileHeader(data[offset:])
            offset = len(file_header)
            yield offset, file_header
        while len(data) - offset > 2:
            token = token_factory(data[offset:], state)
            record = token.parse_token(warn=warn)
            if force:
                record.force()
            offset += len(token)
            yield offset, token
        # if we're here then there are 2 or less bytes remaining
        if len(data) - offset != 2:
            raise Exception('Misaligned checksum')
    except Exception as e:
        raise Exception('Error (%s) at offset %d' % (e, offset))


def slurp(state, data, initial_offset, warn=False, force=True, max_record_len=None):
    '''
    read as much as possible starting at the given offset.

    return offsets of valid tokens and a flag indicating if all data were read.
    '''
    offsets_and_states = []
    offset = initial_offset
    try:
        for offset, token in offset_tokens(state, data, initial_offset, warn=warn, force=force):
            if max_record_len and len(token) > max_record_len:
                log.info('Record too large (%d > %d) at offset %d' % (len(token), max_record_len, offset))
                return offsets_and_states, False
            offsets_and_states.append((offset, state.copy()))
        log.info('Read complete from %d' % initial_offset)
        return offsets_and_states, True
    except Exception as e:
        log.debug(e)
        log.debug('Reading from offset %s found %d tokens before %d' %
                  (initial_offset, len(offsets_and_states), offset))
        return offsets_and_states, False


class Backtrack(Exception): pass


def advance(initial_state, data, drop_count=0, initial_offset=0, warn=False, force=True,
            min_sync_cnt=3, max_record_len=None, max_drop_cnt=1, max_back_cnt=3, max_fwd_len=200):
    '''
    try to synchronize on the stream and then read as much as possible.  if reading later fails,
    then try recurse if not at limit; if that fails, backtrack a record and try again.

    finally return the offsets used.

    "synchronize" means progressively discard data until some number of records smaller than
    some length can be read.
    '''
    initial_offsets_and_states = [(initial_offset, initial_state.copy())]
    offsets_and_states, complete = slurp(initial_state, data, initial_offset,
                                         warn=warn, force=force, max_record_len=max_record_len)
    if offsets_and_states:
        log.debug('%d: Read %d records; offset %d to %d' %
                  (drop_count, len(offsets_and_states), initial_offset, offsets_and_states[-1][0]))
    else:
        log.debug('%d: Did not read any records' % drop_count)
    offsets_and_states = initial_offsets_and_states + offsets_and_states
    if complete:
        # use explicit offset rather than open interval because need to drop final checksum
        if len(offsets_and_states) > 1:
            return [slice( offsets_and_states[0][0], offsets_and_states[-1][0])]
        else:
            return []
    if drop_count and len(offsets_and_states) < min_sync_cnt:  # failing to sync on first read is OK
        raise Backtrack('Failed to sync at offset %d' % initial_offset)
    if drop_count >= max_drop_cnt:
        raise Backtrack('No more drops')
    for delta in range(1, max_fwd_len):
        for back_cnt in range(1, min(max_back_cnt + 1, len(offsets_and_states))):
            offset, state = offsets_and_states[-back_cnt]
            log.debug('Searching forwards from offset %d after dropping %d records' % (offset, back_cnt-1))
            if offset + delta >= len(data):  # > for when a delta of 0 would have done
                log.info('Exhausted data')
                return [slice(initial_offset, offset)]
            else:
                try:
                    log.debug('%d: Retrying (drop %d, skip %d) at offset %d' %
                              (drop_count, back_cnt-1, delta, offset+delta))
                    slices = advance(state.copy(), data, drop_count+1, offset+delta,
                                     min_sync_cnt=min_sync_cnt, max_record_len=max_record_len,
                                     max_drop_cnt=max_drop_cnt, max_fwd_len=max_fwd_len)
                    return [slice(initial_offset, offset)] + slices
                except Backtrack as e:
                    log.debug('%d: Backtrack at (drop %d, skip %d): "%s"' % (drop_count, back_cnt, delta, e))
    raise Backtrack('Search exhausted at %d' % initial_offset)





from logging import getLogger

from .args import PATH, FORMAT, ALL_FIELDS, ALL_MESSAGES, LIMIT_RECORDS, WARN, GREP, MESSAGE, NAME, NOT, MATCH, \
    WIDTH, NO_VALIDATE, INTERNAL, MAX_DELTA_T, AFTER_BYTES, AFTER_RECORDS, CONTEXT, LIMIT_BYTES, COMPACT, FIELD
from ..fit.profile.profile import read_fit
from ..fit.summary import summarize

log = getLogger(__name__)


def fit(args, db):
    '''
## fit

    > ch2 fit PATH [PATH ...]

Print the contents of fit files.

The format and details displayed can be selected with --records,
--tables, --grep, --messages, --fields, and --csv.

For full options see `ch2 fit -h`.

Note: When using bash use `shopt -s globstar` to enable ** globbing.

### Examples

    > ch2 -v 0 fit --records ride.fit

Will print the contents of the file to stdout (use `-v 0` to suppress logging
or redirect stderr elsewhere).

    > ch2 -v 0 fit --grep '.*:sport=cycling' --match 0 --name directory/**/*.fit

Will list file names that contain cycling data.

    > ch2 fit --grep PATTERN -- FILE

You may need a `--` between patterns and file paths so that the argument parser can decide where patterns
finish and paths start.
    '''
    for file_path in args[PATH]:
        summarize(args[FORMAT], read_fit(log, file_path),
                  all_fields=args[ALL_FIELDS], all_messages=args[ALL_MESSAGES], internal=args[INTERNAL],
                  after_bytes=args[AFTER_BYTES], limit_bytes=args[LIMIT_BYTES],
                  after_records=args[AFTER_RECORDS], limit_records=args[LIMIT_RECORDS],
                  messages=args[MESSAGE], fields=args[FIELD], warn=args[WARN], grep=args[GREP],
                  name_file=file_path if args[NAME] else None, invert=args[NOT], match=args[MATCH],
                  compact=args[COMPACT], context=args[CONTEXT], no_validate=args[NO_VALIDATE],
                  max_delta_t=args[MAX_DELTA_T], width=args[WIDTH])



from time import time

from sqlalchemy import Column, Text, Integer

from ..support import Base
from ..types import Time, ShortCls


class SystemConstant(Base):

    __tablename__ = 'system_constant'

    name = Column(Text, primary_key=True)
    value = Column(Text, nullable=False)

    TIMEZONE = 'timezone'


class SystemProcess(Base):

    __tablename__ = 'system_process'

    id = Column(Integer, primary_key=True)
    command = Column(Text, nullable=False)
    log = Column(Text, nullable=False)
    start = Column(Time, nullable=False, default=time)
    owner = Column(ShortCls, nullable=False, index=True)
    pid = Column(Integer, nullable=False, index=True)




import datetime as dt
from collections import namedtuple
from json import loads
from logging import getLogger

from sqlalchemy import desc, inspect, select, and_

from ch2.data import hr_zones
from . import MultiProcCalculator, ActivityJournalCalculatorMixin, DirectCalculatorMixin
from ..load import StatisticJournalLoader
from ..names import FTHR, HR_ZONE, HEART_RATE, S
from ...squeal import Constant, StatisticJournal, StatisticName, StatisticJournalFloat, StatisticJournalInteger, \
    ActivityGroup

log = getLogger(__name__)

# constraint comes from constant
HRImpulse = namedtuple('HRImpulse', 'dest_name, gamma, zero, max_secs')


class HeartRateCalculator(ActivityJournalCalculatorMixin, DirectCalculatorMixin, MultiProcCalculator):

    def __init__(self, *args, cost_calc=5, cost_write=1, impulse=None, **kargs):
        self.impulse = self._assert('impulse', impulse)
        super().__init__(*args, cost_calc=cost_calc, cost_write=cost_write, **kargs)

    def _startup(self, s):
        self.__fthr_cache = {}
        for activity_group in s.query(ActivityGroup).all():
            self.__fthr_cache[activity_group.id] = \
                list(s.query(StatisticJournal).join(StatisticName).
                     filter(StatisticName.name == FTHR,
                            StatisticName.owner == Constant,
                            StatisticName.constraint == activity_group).
                     order_by(desc(StatisticJournal.time)).all())

    def _read_data(self, s, ajournal):
        sn = inspect(StatisticName).local_table
        sj = inspect(StatisticJournal).local_table
        sji = inspect(StatisticJournalInteger).local_table
        stmt = select([sj.c.time, sji.c.value]). \
            select_from(sj.join(sn).join(sji)). \
            where(and_(sj.c.source_id == ajournal.id,
                       sn.c.owner == self.owner_in,
                       sn.c.constraint == ajournal.activity_group,
                       sn.c.name == HEART_RATE)). \
            order_by(sj.c.time)
        # log.debug(stmt)
        return list(s.connection().execute(stmt))

    def _calculate_results(self, s, ajournal, data, loader):

        hr_impulse = HRImpulse(**loads(Constant.get(s, self.impulse).at(s).value))
        log.debug('%s: %s' % (self.impulse, hr_impulse))
        impulses = []
        prev_time, prev_heart_rate_zone = None, None

        for time, heart_rate in data:
            if heart_rate:
                heart_rate_zone = self._calculate_zone(s, heart_rate, time, ajournal.activity_group)
                if heart_rate_zone is not None:
                    loader.add(HR_ZONE, None, None, ajournal.activity_group, ajournal, heart_rate_zone, time,
                               StatisticJournalFloat)
            else:
                heart_rate_zone = None
            if prev_heart_rate_zone is not None:
                duration = (time - prev_time).total_seconds()
                if duration <= hr_impulse.max_secs:
                    heart_rate_impulse = self._calculate_impulse(prev_heart_rate_zone, duration, hr_impulse)
                    impulses.append((heart_rate_impulse, time))
                    loader.add(hr_impulse.dest_name, None, None, ajournal.activity_group, ajournal,
                               heart_rate_impulse, time, StatisticJournalFloat)
                    loader.add('%s (duration)' % hr_impulse.dest_name, S, None, ajournal.activity_group, ajournal,
                               duration, time, StatisticJournalFloat)
            elif not impulses:
                impulses.append((0, time))
            prev_time, prev_heart_rate_zone = time, heart_rate_zone

        if impulses:
            loader = StatisticJournalLoader(s, self)
            self._interpolate(hr_impulse.dest_name, loader, impulses, ajournal)
            # if there are no values, add a single null so we don't re-process
            if not loader:
                loader.add(HR_ZONE, None, None, ajournal.activity_group, ajournal, None, ajournal.start,
                           StatisticJournalFloat)
            loader.load()

    def _calculate_impulse(self, heart_rate_zone, duration, hr_impulse):
        return duration * ((max(heart_rate_zone, hr_impulse.zero) - hr_impulse.zero)
                           / (6 - hr_impulse.zero)) ** hr_impulse.gamma

    def _calculate_zone(self, s, heart_rate, time, activity_group):
        for fthr in self.__fthr_cache[activity_group.id]:
            if fthr.time <= time:  # ordered in descending time
                lower_limit, prev_delta = 0, None
                zones = hr_zones(fthr.value)
                for zone, upper_limit in enumerate(zones):
                    if lower_limit <= heart_rate < upper_limit:
                        if zone == 0:
                            return 1 + zone
                        elif zone == 5:
                            return 1 + zone + (heart_rate - lower_limit) / prev_delta
                        else:
                            return 1 + zone + (heart_rate - lower_limit) / (upper_limit - lower_limit)
                    prev_delta = upper_limit - lower_limit
                    lower_limit = upper_limit

    def _interpolate(self, name, loader, impulses, ajournal, interval=10):

        # we need evenly-sampled statistics so we can do distributions over time.
        # why interpolate just this one statistic?
        # because it's not the same as others - it's already time-based
        # so interpolating it later is more complex.
        # but i may change my mind

        # you can configure names, but plotting assumes HR_IMPULSE_10 exists...

        integral, sum = [], 0
        for impulse, time in impulses:
            sum += impulse
            integral.append((sum, time))

        impulse_0, time_0 = integral.pop(0)
        time, prev = time_0, 0
        while integral:
            impulse_1, time_1 = integral[0]
            delta = (time_1 - time_0).total_seconds()
            while time <= time_1:
                interp = impulse_0 + (impulse_1 - impulse_0) * (time - time_0).total_seconds() / delta
                diff = (interp - prev) / interval
                loader.add(f'{name} / {interval}s', None, None, ajournal.activity_group, ajournal, diff, time,
                           StatisticJournalFloat)
                time += dt.timedelta(seconds=interval)
                prev = interp
            while integral and time > integral[0][1]:
                impulse_0, time_0 = integral.pop(0)




from logging import getLogger, NullHandler
from sys import version_info

getLogger('bokeh').addHandler(NullHandler())
getLogger('tornado').addHandler(NullHandler())

from .commands.activities import activities
from .commands.args import COMMAND, parser, NamespaceWithVariables, PROGNAME, HELP, DEV, DIARY, FIT, \
    PACKAGE_FIT_PROFILE, ACTIVITIES, NO_OP, DEFAULT_CONFIG, CONSTANTS, STATISTICS, TEST_SCHEDULE, MONITOR, GARMIN, \
    UNLOCK, DUMP, FIX_FIT, CH2_VERSION
from .commands.constants import constants
from .commands.dump import dump
from .commands.default_config import default_config
from .commands.diary import diary
from .commands.fit import fit
from .commands.fix_fit import fix_fit
from .commands.garmin import garmin
from .commands.help import help, LengthFmt
from .commands.monitor import monitor
from .commands.package_fit_profile import package_fit_profile
from .commands.statistics import statistics
from .commands.test_schedule import test_schedule
from .commands.unlock import unlock
from .lib.io import tui
from .lib.log import make_log, log_current_exception
from .squeal.database import Database
from .squeal import SystemConstant
from .uranus.server import set_jupyter_args, stop_jupyter

log = getLogger(__name__)


@tui
def no_op(args, db):
    '''
## no-op

This is used internally when accessing data in Jupyter or configuring the system
at the command line.
    '''
    pass


COMMANDS = {ACTIVITIES: activities,
            CONSTANTS: constants,
            DEFAULT_CONFIG: default_config,
            DIARY: diary,
            DUMP: dump,
            FIT: fit,
            FIX_FIT: fix_fit,
            GARMIN: garmin,
            HELP: help,
            MONITOR: monitor,
            STATISTICS: statistics,
            NO_OP: no_op,
            PACKAGE_FIT_PROFILE: package_fit_profile,
            TEST_SCHEDULE: test_schedule,
            UNLOCK: unlock,
            }


def main():
    args = NamespaceWithVariables(parser().parse_args())
    command_name = args[COMMAND] if COMMAND in args else None
    command = COMMANDS[command_name] if command_name in COMMANDS else None
    tui = command and hasattr(command, 'tui') and command.tui
    make_log(args, tui=tui)
    log.info('Version %s' % CH2_VERSION)
    if version_info < (3, 7):
        raise Exception('Please user Python 3.7 or more recent')
    db = Database(args, log)
    try:
        if db.is_empty() and (not command or command_name != DEFAULT_CONFIG):
            refuse_until_configured()
        else:
            set_jupyter_args(log, args)
            try:
                if command:
                    command(args, db)
                else:
                    log.debug('If you are seeing the "No command given" error during development ' +
                              'you may have forgotten to set the command name via `set_defaults()`.')
                    raise Exception('No command given (try `ch2 help`)')
            finally:
                stop_jupyter(log)
    except KeyboardInterrupt:
        log.critical('User abort')
        exit(1)
    except Exception as e:
        log.critical(e)
        log_current_exception()
        log.info('See `%s %s` for available commands.' % (PROGNAME, HELP))
        log.info('Docs at http://andrewcooke.github.io/choochoo')
        if not args or args[DEV]:
            raise
        exit(2)


def refuse_until_configured():
    LengthFmt().print_all('''
Welcome to Choochoo.

Before using the ch2 command you must configure the system.

Please see the documentation at http://andrewcooke.github.io/choochoo

To generate a default configuration use the command

    %s %s

NOTE: The default configuration is only an example.  Please see the docs
for more details.''' % (PROGNAME, DEFAULT_CONFIG))




from .tables.activity import ActivityGroup, ActivityTimespan, ActivityJournal, ActivityBookmark
from .tables.constant import Constant
from .tables.system import SystemConstant, SystemProcess
from .tables.monitor import MonitorJournal
from .tables.nearby import ActivitySimilarity, ActivityNearby
from .tables.pipeline import Pipeline, PipelineType
from .tables.segment import Segment, SegmentJournal
from .tables.source import Source, Interval, NoStatistics, Dummy, Composite, CompositeComponent
from .tables.statistic import StatisticName, StatisticJournalFloat, StatisticJournalText, StatisticJournalInteger, \
    StatisticJournal, StatisticMeasure, StatisticJournalType
from .tables.topic import TopicJournal, Topic, TopicField
from .tables.timestamp import Timestamp




from logging import getLogger
from re import search

from urwid import Text, Pile, Columns, Divider

from . import JournalDiary
from .heart_rate import build_zones
from ..calculate.activity import ActivityCalculator, MAX_MEAN_EP_M_ANY
from ..calculate.power import PowerCalculator
from ..names import ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED, MED_KM_TIME_ANY, MAX_MED_HR_M_ANY, CLIMB_ELEVATION, \
    CLIMB_DISTANCE, CLIMB_GRADIENT, CLIMB_TIME, TOTAL_CLIMB, MIN_KM_TIME_ANY, MAX_MED_EP_M_ANY, CALORIE_ESTIMATE, \
    ENERGY_ESTIMATE
from ...data.climb import climbs_for_activity
from ...lib.date import format_seconds, time_to_local_time, to_time, HMS
from ...lib.utils import label
from ...squeal.tables.statistic import StatisticJournal, StatisticName
from ...uweird.fields.summary import summary_columns
from ...uweird.tui.decorators import Indent

log = getLogger(__name__)
HRZ_WIDTH = 30


class ActivityDiary(JournalDiary):

    def _journal_date(self, s, f, ajournal, date):
        zones = build_zones(s, ajournal, HRZ_WIDTH)
        active_date = self.__active_date(s, ajournal, date)
        climbs = self.__climbs(s, ajournal, date)
        details = Pile(([] if climbs else [Divider()]) + active_date + climbs)
        yield Pile([
            Text(ajournal.name),
            Indent(Columns([details, (HRZ_WIDTH + 2, zones)])),
            Divider(),
            Indent(Columns([Pile(self.__template(s, ajournal, MIN_KM_TIME_ANY, 'Min Time', r'(\d+km)', date) +
                                 self.__template(s, ajournal, MED_KM_TIME_ANY, 'Med Time', r'(\d+km)', date)),
                            Pile(self.__template(s, ajournal, MAX_MED_HR_M_ANY, 'Max Med HR', r'(\d+m)', date) +
                                 self.__template(s, ajournal, MAX_MEAN_EP_M_ANY, 'Max Mean EP', r'(\d+m)', date))]))
        ])

    def __active_date(self, s, ajournal, date):
        body = []
        body.append(Text('%s - %s  (%s)' % (time_to_local_time(to_time(ajournal.start)),
                                            time_to_local_time(to_time(ajournal.finish), fmt=HMS),
                                            format_seconds((ajournal.finish - ajournal.start).seconds))))
        for name in (ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED):
            sjournal = StatisticJournal.at(s, ajournal.start, name, ActivityCalculator, ajournal.activity_group)
            if sjournal:
                body.append(Text([label('%s: ' % sjournal.statistic_name.name)] + self.__format_value(sjournal, date)))
        for name in (ENERGY_ESTIMATE, CALORIE_ESTIMATE,):
            sjournal = StatisticJournal.at(s, ajournal.start, name, PowerCalculator, ajournal.activity_group)
            if sjournal:
                body.append(Text([label('%s: ' % sjournal.statistic_name.name)] + self.__format_value(sjournal, date)))
        return body

    def __climbs(self, s, ajournal, date):
        total, climbs = climbs_for_activity(s, ajournal)
        if total:
            body = [Text(['Climbs ', label('Total: '), '%dm ' % total.value, *total.measures_as_text(date)])]
            for climb in climbs:
                body.append(Text(['%3sm/%.1fkm (%d%%)' %
                                  (int(climb[CLIMB_ELEVATION].value), # display int() with %s to get space padding
                                   climb[CLIMB_DISTANCE].value / 1000, climb[CLIMB_GRADIENT].value),
                                  label(' in '),
                                  format_seconds(climb[CLIMB_TIME].value),
                                  ' ',
                                  *climb[CLIMB_ELEVATION].measures_as_text(date),
                                  ]))
            return body
        else:
            return []

    def __template(self, s, ajournal, template, title, re, date):
        body = [Text(title)]
        sjournals = s.query(StatisticJournal).join(StatisticName). \
            filter(StatisticJournal.time == ajournal.start,
                   StatisticName.name.like(template),
                   StatisticName.owner == ActivityCalculator,
                   StatisticName.constraint == ajournal.activity_group).order_by(StatisticName.name).all()
        # extract
        for sjournal in self.__sort_journals(sjournals):
            body.append(Text([label(search(re, sjournal.statistic_name.name).group(1) + ': ')] +
                             self.__format_value(sjournal, date)))
        return body

    def __sort_journals(self, sjournals):
        return sorted(sjournals,
                      # order by integer embedded in name
                      key=lambda sjournal: int(search(r'(\d+)', sjournal.statistic_name.name).group(1)))

    def __sort_names(self, statistic_names):
        return sorted(statistic_names,
                      # order by integer embedded in name
                      key=lambda statistic_name: int(search(r'(\d+)', statistic_name.name).group(1)))

    def __format_value(self, sjournal, date):
        return ['%s ' % sjournal.formatted()] + sjournal.measures_as_text(date)

    def _display_schedule(self, s, f, date, schedule=None):
        columns = list(self.__schedule_fields(s, f, date, schedule))
        if columns:
            yield Pile([Text('Activities'),
                        Indent(Pile(columns))])

    def __schedule_fields(self, s, f, date, schedule):
        names = list(self.__names(s, ACTIVE_DISTANCE, ACTIVE_TIME, ACTIVE_SPEED,
                                  TOTAL_CLIMB, CLIMB_ELEVATION, CLIMB_DISTANCE, CLIMB_GRADIENT, CLIMB_TIME))
        yield from summary_columns(log, s, f, date, schedule, names)
        names = self.__sort_names(self.__names_like(s, MIN_KM_TIME_ANY))
        yield from summary_columns(log, s, f, date, schedule, names)
        names = self.__sort_names(self.__names_like(s, MED_KM_TIME_ANY))
        yield from summary_columns(log, s, f, date, schedule, names)
        names = self.__sort_names(self.__names_like(s, MAX_MED_HR_M_ANY))
        yield from summary_columns(log, s, f, date, schedule, names)

    def __names(self, s, *names):
        for name in names:
            yield s.query(StatisticName). \
                filter(StatisticName.name == name,
                       StatisticName.owner == ActivityCalculator).one()

    def __names_like(self, s, name):
        return s.query(StatisticName). \
            filter(StatisticName.name.like(name),
                   StatisticName.owner == ACTIVE_DISTANCE).all()




from .date import local_date_to_time, to_date, time_to_local_time, format_date, format_seconds, format_time, \
    local_time_to_time, time_to_local_date, to_duration, to_time
from .utils import short_str, sign, datetime_to_epoch, format_metres, format_percent





import datetime as dt
from abc import abstractmethod
from logging import getLogger

from sqlalchemy import or_
from urwid import MainLoop, Columns, Pile, Frame, Filler, Text, Divider, WEIGHT, connect_signal, Padding

from .args import DATE, SCHEDULE, FAST
from ..lib.date import to_date, YMD
from ..lib.io import tui
from ..lib.schedule import Schedule
from ..lib.utils import PALETTE_RAINBOW, em, label
from ..lib.widgets import DateSwitcher
from ..squeal import PipelineType, Topic, TopicJournal
from ..squeal.database import ActivityJournal, StatisticJournal
from ch2.squeal.utils import add
from ..stoats.display import display_pipeline
from ..stoats.display.nearby import nearby_any_time, fmt_nearby
from ..stoats.pipeline import run_pipeline
from ..uranus.template.activity_details import activity_details
from ..uranus.template.all_activities import all_activities
from ..uranus.template.compare_activities import compare_activities
from ..uranus.template.health import health
from ..uranus.template.similar_activities import similar_activities
from ..uweird.fields import PAGE_WIDTH
from ..uweird.fields.summary import summary_columns
from ..uweird.tui.decorators import Border, Indent
from ..uweird.tui.factory import Factory
from ..uweird.tui.fixed import Fixed
from ..uweird.tui.tabs import TabList
from ..uweird.tui.widgets import DividedPile, ArrowMenu, SquareButton

log = getLogger(__name__)

@tui
def diary(args, db):
    '''
## diary

    > ch2 diary [DATE]

The date can be an absolute day or the number of days previous.  So `ch2 diary 1` selects yesterday.

Display the daily diary.  Enter information here.

To exit, alt-q (or, without saving, alt-x).

    > ch2 diary (--month | --year | --schedule SCHEDULE) [DATE}

Display a summary for the month / year / schedule.
    '''
    date, schedule = args[DATE], args[SCHEDULE]
    if not date:
        date = dt.date.today()
    else:
        days = None
        try:
            # try int first because we need to separate the case of days from years
            days = int(date)
            if days > 1000:
                days = None
        except ValueError:
            pass
        if days is None:
            date = to_date(date)
        else:
            date = dt.date.today() - dt.timedelta(days=days)
    with db.session_context() as s:
        TopicJournal.check_tz(log, s)
    if schedule:
        schedule = Schedule(schedule)
        if schedule.start or schedule.finish:
            raise Exception('Schedule must be open (no start or finish)')
        MainLoop(ScheduleDiary(log, db, date, schedule), palette=PALETTE_RAINBOW).run()
    else:
        MainLoop(DailyDiary(log, db, date), palette=PALETTE_RAINBOW).run()
        if not args[FAST]:
            print('\n  Please wait while statistics are updated...')
            run_pipeline(db, PipelineType.STATISTIC)
            print('  ...done (thanks!)\n')


class Diary(DateSwitcher):

    def __init__(self, log, db, date):
        super().__init__(log, db, date)

    def save(self):
        s = self._session
        self._log.info(f'{bool(s)}')
        if s:
            self.__clean(s, s.dirty, delete=True)
            self.__clean(s, s.new, delete=False)
        super().save()

    def __clean(self, s, instances, delete=False):
        self._log.debug(f'{len(instances)}')
        for instance in instances:
            if isinstance(instance, StatisticJournal):
                if instance.value == None:
                    self._log.debug(f'Discarding {instance}')
                    if delete:
                        s.delete(instance)
                    else:
                        s.expunge(instance)

    def _build(self, s):
        self._log.debug('Building diary at %s' % self._date)
        body, f = [], Factory(TabList())
        root_topics = list(self._topics(s))
        for topic in root_topics:
            body.append(self.__display_topic(s, f, topic))
        for extra in self._display_pipeline(s, f):
            body.append(extra)
        for extra in self._display_gui(s, f):
            body.append(extra)
        self._check_body(body)
        body = Border(Frame(Filler(DividedPile(body), valign='top'),
                            header=Pile([self._header(), Divider()]),
                            footer=Pile([Divider(), Text(self.__footer(), align='center')])))
        return body, f.tabs

    @abstractmethod
    def _header(self):
        pass

    def _check_body(self, body):
        pass

    def __footer(self):
        footer = ['meta-', em('q'), 'uit/e', em('x'), 'it/', em('s'), 'ave']
        footer += [' [shift]meta-']
        for sep, c in enumerate('dwmya'):
            if sep: footer += ['/']
            footer += [em(c)]
        footer += ['ctivity/', em('t'), 'oday']
        return footer

    def __display_topic(self, s, f, topic):
        self._log.debug('%s' % topic)
        body, title = [], None
        if topic.name:
            title = Text(topic.name)
        if topic.description:
            body.append(Text(topic.description))
        body += list(self._display_fields(s, f, topic))
        body += list(self.__display_children(s, f, topic))
        if not body:
            return title
        body = Indent(Pile(body))
        if title:
            body = Pile([title, body])
        return body

    def __display_children(self, s, f, topic):
        for child in topic.children:
            if self._filter_child(child):
                extra = self.__display_topic(s, f, child)
                if extra:
                    yield extra

    def _filter_child(self, child):
        return child.schedule.at_location(self._date)

    def _display_gui(self, s, f):
        yield from []


class DailyDiary(Diary):
    '''
    Render the diary at a given date.
    '''

    def _topics(self, s):
        for topic in s.query(Topic).filter(Topic.parent == None,
                                           or_(Topic.start <= self._date, Topic.start == None),
                                           or_(Topic.finish >= self._date, Topic.finish == None)). \
                order_by(Topic.sort).all():
            if topic.schedule.at_location(self._date):
                yield topic

    def _header(self):
        return Text(self._date.strftime('%Y-%m-%d - %A'))

    def _display_fields(self, s, f, topic):
        columns, width = [], 0
        tjournal = self.__topic_journal(s, topic)
        tjournal.populate(self._log, s)
        for field in topic.fields:
            if field in tjournal.statistics:  # might be outside schedule
                self._log.debug('%s' % field)
                display = field.display_cls(self._log, tjournal.statistics[field],
                                            *field.display_args, **field.display_kargs)
                self._log.debug('%s' % display)
                if width + display.width > PAGE_WIDTH:
                    yield Columns(columns)
                    columns, width = [], 0
                columns.append((WEIGHT, display.width, f(display.widget())))
                width += display.width
        if width:
            yield Columns(columns)

    def __topic_journal(self, s, topic):
        tjournal = s.query(TopicJournal). \
            filter(TopicJournal.topic == topic,
                   TopicJournal.date == self._date).one_or_none()
        if not tjournal:
            tjournal = add(s, TopicJournal(topic=topic, date=self._date))
        return tjournal

    def _display_pipeline(self, s, f):
        yield from display_pipeline(s, f, self._date, self)

    def _display_gui(self, s, f):
        menus = list(self.__gui_menus(s, f))
        if menus:
            yield Pile([Text('Jupyter'), Indent(Pile(menus))])

    def __gui_menus(self, s, f):
        for aj1 in ActivityJournal.at_date(s, self._date):
            options = [(None, 'None')] + [(aj2, fmt_nearby(aj2, nb)) for aj2, nb in nearby_any_time(s, aj1)]
            menu = ArrowMenu(label('%s v ' % aj1.name), dict(options))
            connect_signal(menu, 'click', self.__show_gui, aj1)
            button = SquareButton('All Similar')
            connect_signal(button, 'click', self.__show_similar, aj1)
            yield Columns([f(menu), f(Padding(Fixed(button, 13), width='clip'))])
        button = SquareButton('Health')
        connect_signal(button, 'click', self.__show_health, self._date)
        yield f(Padding(Fixed(button, 8), width='clip'))

    def __show_gui(self, w, aj1):
        if w.state:
            compare_activities(aj1.start, w.state.start, aj1.activity_group.name, log=self._log)
        else:
            activity_details(aj1.start, aj1.activity_group.name, log=self._log)

    def __show_similar(self, w, aj1):
        similar_activities(aj1.start, aj1.activity_group.name, log=self._log)

    def __show_health(self, w, date):
        health(date.strftime(YMD), log=self._log)


class ScheduleDiary(Diary):
    '''
    Display summary data for the given schedule.
    '''

    def __init__(self, log, db, date, schedule):
        self._schedule = schedule
        super().__init__(log, db, self._refine_new_date(date))

    def _header(self):
        return Text(self._date.strftime('%Y-%m-%d') + ' - Summary for %s' % self._schedule.describe())

    def _check_body(self, body):
        if len(body) < 4:
            body.append(Indent(Text('Updating the database automatically deletes summary statistics that cover '
                                    + 'the modified data.  You probably need to re-generate the statistics by '
                                    + 'running `ch2 statistics`.')))

    def _refine_new_date(self, date):
        return self._schedule.start_of_frame(date)

    def _topics(self, s):
        finish = self._schedule.next_frame(self._date)
        return s.query(Topic).filter(Topic.parent == None,
                                     or_(Topic.start < finish, Topic.start == None),
                                     or_(Topic.finish >= self._date, Topic.finish == None)). \
            order_by(Topic.sort).all()

    def _filter_child(self, child):
        finish = self._schedule.next_frame(self._date)
        return (child.start is None or child.start < finish) and (child.finish is None or child.finish > self._date)

    def _display_fields(self, s, f, topic):
        names = [field.statistic_name for field in topic.fields]
        yield from summary_columns(self._log, s, f, self._date, self._schedule, names)

    def _display_pipeline(self, s, f):
        yield from display_pipeline(s, f, self._date, self, schedule=self._schedule)

    def _display_gui(self, s, f):
        button = SquareButton('All Activities')
        connect_signal(button, 'click', self.__show_all)
        yield Pile([Text('Jupyter'), Indent(f(Padding(Fixed(button, 16), width='clip')))])

    def __show_all(self, w):
        finish = self._schedule.next_frame(self._date)
        all_activities(self._date.strftime('%Y-%m-%d'), finish.strftime('%Y-%m-%d'), log=self._log)




from json import dumps

from . import name_constant, add_enum_constant, set_constant, add_statistics
from ..stoats.calculate.power import Power, ExtendedPowerCalculator

POWER_ESTIMATE_CNAME = 'PowerEstimate'


def add_power_estimate(s, c, activity_group,
                       bike='#$SegmentReader:Bike',
                       rider_weight='$Topic:Weight:Topic \"Diary\" (d)',
                       vary='wind_speed, wind_heading, slope'):
    '''
    Add the constants necessary to estimate power output.
    '''
    activity_group_constraint = str(activity_group)
    power_name = name_constant(POWER_ESTIMATE_CNAME, activity_group)
    power = add_enum_constant(s, power_name, Power, single=False, constraint=activity_group_constraint,
                              description='Data needed to estimate power - see Power enum')
    set_constant(s, power, dumps({'bike': bike, 'rider_weight': rider_weight, 'vary': vary}))
    add_statistics(s, ExtendedPowerCalculator, c, owner_in='[unused - data via activity_statistics]',
                   power=name_constant(POWER_ESTIMATE_CNAME, activity_group))




from bokeh.plotting import output_notebook, show

from ch2.data import *
from ch2.uranus.decorator import template


@template
def similar_activities(local_time, activity_group_name):

    f'''
    # Similar Activities: {local_time.split()[0]}
    '''

    '''
    $contents
    '''

    '''
    ## Build Maps
    
    Loop over activities, retrieve data, and construct maps. 
    '''

    s = session('-v2')

    maps = [map_thumbnail(100, 120, data)
            for data in (activity_statistics(s, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y,
                                             activity_journal_id=similar[0].id).resample('1min').mean()
                         for similar in nearby_activities(s, local_time=local_time,
                                                          activity_group_name=activity_group_name))
            if len(data.dropna()) > 10]

    print(f'Found {len(maps)} activities')

    '''
    ## Display Maps
    '''

    output_notebook()
    show(tile(maps, 8))




from logging import getLogger
from os import getpid
from subprocess import Popen
from sys import argv
from time import sleep, mktime, time

from psutil import pid_exists, Process

from ..squeal import SystemProcess
from ..squeal.types import short_cls


log = getLogger(__name__)
DELTA_TIME = 3
SLEEP_TIME = 1
REPORT_TIME = 60
LOG = 'log'
LIKE = 'like'


class Workers:

    def __init__(self, s, n_parallel, owner, cmd):
        self._s = s
        self.n_parallel = n_parallel
        self.owner = owner
        self.cmd = cmd
        self.__workers = {}  # map from Popen to log index
        self.ch2 = self.__get_ch2()
        self.clear_all()

    def __get_ch2(self):
        with open(f'/proc/{getpid()}/cmdline', 'rb') as f:
            line = f.readline()

            def parse():
                word = bytearray()
                for char in line:
                    if char:
                        word.append(char)
                    else:
                        yield word.decode('utf8')
                        word = bytearray()

            words = list(parse())
            if len(argv) > 1:
                i = words.index(argv[1])
                words = words[:i]
            ch2 = ' '.join(words)
            log.debug(f'Using command "{ch2}"')
            return ch2

    def clear_all(self):
        for worker in self.__workers.keys():
            log.warning(f'Killing PID {worker.pid} ({worker.args})')
            self._delete_pid(worker.pid)
            worker.kill()
        for process in self._s.query(SystemProcess). \
                filter(SystemProcess.owner == self.owner).all():
            if pid_exists(process.pid):
                p = Process(process.pid)
                if abs(p.create_time() - mktime(process.start.timetuple())) < DELTA_TIME:
                    log.warning(f'Killing rogue PID {process.pid} ({" ".join(p.cmdline())})')
                    p.kill()
            self._s.delete(process)
        self._s.commit()

    def _delete_pid(self, pid):
        # weird sqlalchemy bug(?) here.  calling .delete() on the query causes an inconsistent session
        # so retrieve and then delete from the session.
        # may not be a bug - may be related to synchronize_session.
        # could perhaps avoid the extra step by specifying that.
        p = self._s.query(SystemProcess). \
            filter(SystemProcess.owner == self.owner,
                   SystemProcess.pid == pid).one()
        self._s.delete(p)
        self._s.commit()

    def _read_pid(self, pid):
        return self._s.query(SystemProcess). \
            filter(SystemProcess.owner == self.owner,
                   SystemProcess.pid == pid).one()

    def run(self, args):
        self.wait(self.n_parallel - 1)
        log_index = self._free_log_index()
        log_name = f'{short_cls(self.owner)}.{log_index}.{LOG}'
        cmd = (self.cmd + ' ' + args).format(log=log_name, ch2=self.ch2)
        worker = Popen(args=cmd, shell=True)
        log.debug(f'Adding command [{cmd}] (PID {worker.pid})')
        self._s.add(SystemProcess(command=cmd, owner=self.owner, pid=worker.pid, log=log_name))
        self._s.commit()  # critical, or database is locked for worker
        self.__workers[worker] = log_index

    def wait(self, n_workers=0):
        last_report = 0
        while len(self.__workers) > n_workers:
            if time() - last_report > REPORT_TIME:
                log.debug(f'Currently have {len(self.__workers)} workers; waiting to drop to {n_workers}')
                last_report = time()
            for worker in list(self.__workers.keys()):
                worker.poll()
                process = self._read_pid(worker.pid)
                if worker.returncode is not None:
                    if worker.returncode:
                        msg = f'Command "{process.command}" exited with return code {worker.returncode} ' + \
                              f'see {process.log} for more info'
                        log.warning(msg)
                        self.clear_all()
                        raise Exception(msg)
                    else:
                        log.debug(f'Command "{process.command}" finished successfully')
                        del self.__workers[worker]
                        self._delete_pid(worker.pid)
            sleep(SLEEP_TIME)
        if last_report:
            log.debug(f'Now have {len(self.__workers)} workers')

    def _free_log_index(self):
        used = set(self.__workers.values())
        for i in range(self.n_parallel):
            if i not in used:
                return i
        raise Exception('No log available (too many workers)')




from sqlalchemy import or_, desc, distinct
from sqlalchemy.orm import aliased
from sqlalchemy.sql.functions import min
from urwid import Pile, Text, Columns, connect_signal

from . import JournalDiary
from ...lib.date import to_time, time_to_local_date
from ...lib.utils import label
from ...squeal import ActivityJournal, ActivitySimilarity, ActivityNearby
from ...uweird.tui.decorators import Indent
from ...uweird.tui.widgets import ArrowMenu


def _fmt_time(time):
    return to_time(time).strftime('%Y-%m-%d')


def fmt_nearby(aj, nb):
    return to_time(aj.start).strftime('%y-%m-%d') + ' %d%%' % (nb.similarity * 100)


class NearbyDiary(JournalDiary):

    def _display_schedule(self, s, f, date, schedule=None):
        yield from []

    def _journal_date(self, s, f, ajournal, date):
        for constraint in constraints(s):
            row = []
            row += self._any_time(s, f, ajournal, constraint)
            row += self._earlier(s, f, ajournal, constraint)
            row += self._group(s, f, ajournal, constraint)
            if row:
                yield Pile([Text(constraint),
                            Indent(Columns(row))])

    def __change_date(self, w):
        self._diary._change_date(time_to_local_date(w.state))

    def __button(self, f, caption, options):
        if options:
            menu = ArrowMenu(label(caption), options)
            connect_signal(menu, 'click', self.__change_date)
            yield f(menu)

    def _any_time(self, s, f, ajournal, constraint):
        yield from self.__button(f, 'Any Time: ',
                                 dict((aj.start, fmt_nearby(aj, nb)) for aj, nb in
                                      nearby_any_time(s, ajournal, constraint=constraint)))

    def _earlier(self, s, f, ajournal, constraint):
        yield from self.__button(f, 'Earlier: ',
                                 dict((aj.start, fmt_nearby(aj, nb)) for aj, nb in
                                      nearby_earlier(s, ajournal, constraint=constraint)))

    def _group(self, s, f, ajournal, constraint):
        yield from self.__button(f, 'Group: ',
                                 dict((aj.start, _fmt_time(aj.start)) for aj in group(s, ajournal, constraint)))


def constraints(s):
    yield from (c[0] for c in
                s.query(distinct(ActivitySimilarity.constraint)).
                    order_by(ActivitySimilarity.constraint).all())


def group(s, ajournal, constraint):
    nb_us = aliased(ActivityNearby)
    nb_them = aliased(ActivityNearby)
    return [nb.activity_journal
            for nb in s.query(nb_them).
                join(nb_us, nb_us.group == nb_them.group).
                join(ActivityJournal, ActivityJournal.id == nb_them.activity_journal_id).
                filter(nb_us.constraint == constraint,
                       nb_us.activity_journal == ajournal,
                       nb_them.activity_journal != ajournal).
                order_by(ActivityJournal.start)]


def nearby_earlier(s, ajournal, constraint=None, threshold=0.3):
    ajlo = aliased(ActivityJournal)
    ajhi = aliased(ActivityJournal)
    q = s.query(ActivitySimilarity). \
                join(ajhi, ActivitySimilarity.activity_journal_hi_id == ajhi.id). \
                join(ajlo, ActivitySimilarity.activity_journal_lo_id == ajlo.id). \
                filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                           ActivitySimilarity.activity_journal_lo_id == ajournal.id),
                       or_(ajhi.id == ajournal.id, ajhi.start < ajournal.start),
                       or_(ajhi.id == ajournal.id, ajlo.start < ajournal.start),
                       ActivitySimilarity.similarity > threshold). \
                order_by(desc(min(ajlo.start, ajhi.start)))
    if constraint:
        q = q.filter(ActivitySimilarity.constraint == constraint)
    return [(asm.activity_journal_lo if asm.activity_journal_lo != ajournal else asm.activity_journal_hi, asm)
            for asm in q.all()]


def nearby_any_time(s, ajournal, constraint=None, threshold=0.3):
    q = s.query(ActivitySimilarity). \
                filter(or_(ActivitySimilarity.activity_journal_hi_id == ajournal.id,
                           ActivitySimilarity.activity_journal_lo_id == ajournal.id),
                       ActivitySimilarity.similarity > threshold). \
                order_by(desc(ActivitySimilarity.similarity))
    if constraint:
        q = q.filter(ActivitySimilarity.constraint == constraint)
    return [(asm.activity_journal_lo if asm.activity_journal_lo != ajournal else asm.activity_journal_hi, asm)
            for asm in q.all()]




from logging import getLogger
from os.path import splitext, basename

from pygeotile.point import Point

from ..names import LATITUDE, LONGITUDE, M, SPHERICAL_MERCATOR_X, SPHERICAL_MERCATOR_Y, ELEVATION, RAW_ELEVATION
from ..read import AbortImport, MultiProcFitReader
from ...commands.args import ACTIVITIES, WORKER, FAST, mm, FORCE
from ...fit.format.records import fix_degrees, merge_duplicates, no_bad_values
from ...lib.date import to_time
from ...sortem.bilinear import bilinear_elevation_from_constant
from ...squeal.database import Timestamp, StatisticJournalText
from ...squeal.tables.activity import ActivityGroup, ActivityJournal, ActivityTimespan
from ...squeal.tables.statistic import StatisticJournalFloat, STATISTIC_JOURNAL_CLASSES
from ...squeal.utils import add

log = getLogger(__name__)


class ActivityReader(MultiProcFitReader):

    def __init__(self, *args, cost_calc=4, cost_write=1, constants=None, sport_to_activity=None, record_to_db=None,
                 **kargs):
        self.constants = constants
        self.sport_to_activity = self._assert('sport_to_activity', sport_to_activity)
        self.record_to_db = [(field, name, units, STATISTIC_JOURNAL_CLASSES[type])
                             for field, (name, units, type)
                             in self._assert('record_to_db', record_to_db).items()]
        self.add_elevation = not any(name == ELEVATION for (field, name, units, type) in self.record_to_db)
        super().__init__(*args, cost_calc=cost_calc, cost_write=cost_write, **kargs)

    def _base_command(self):
        if self.constants:
            constants = ' '.join(f'-D "{constant}"' for constant in self.constants) + ' -- '
        else:
            constants = ''
        return f'{{ch2}} -v0 -l {{log}} {ACTIVITIES} {mm(WORKER)} {self.id} ' \
            f'{mm(FAST)} {mm(FORCE) if self.force else ""} {constants}'

    def _startup(self, s):
        super()._startup(s)
        self.__oracle = bilinear_elevation_from_constant(log, s)

    def _read_data(self, s, path):
        log.info('Reading activity data from %s' % path)
        records = self._read_fit_file(path, merge_duplicates, fix_degrees, no_bad_values)
        ajournal, activity_group, first_timestamp = self._create_activity(s, path, records)
        self._load_constants(s, ajournal)
        return ajournal.id, (ajournal, activity_group, first_timestamp, path, records)

    def _create_activity(self, s, path, records):
        first_timestamp = self._first(path, records, 'event', 'record').value.timestamp
        sport = self._first(path, records, 'sport').value.sport.lower()
        activity_group = self._activity_group(s, path, sport)
        self._delete_journals(s, activity_group, first_timestamp)
        ajournal = add(s, ActivityJournal(activity_group=activity_group,
                                          start=first_timestamp, finish=first_timestamp,  # will be over-written later
                                          fit_file=path, name=splitext(basename(path))[0]))
        return ajournal, activity_group, first_timestamp

    def _activity_group(self, s, path, sport):
        if sport in self.sport_to_activity:
            return self._lookup_activity_group(s, self.sport_to_activity[sport])
        else:
            log.warning('Unrecognised sport: "%s" in %s' % (sport, path))
            raise AbortImport()

    def _lookup_activity_group(self, s, name):
        activity_group = s.query(ActivityGroup).filter(ActivityGroup.name == name).one_or_none()
        if not activity_group:
            activities = s.query(ActivityGroup).all()
            if activities:
                log.info('Available activity group:')
                for activity_group in activities:
                    log.info('%s - %s' % (activity_group.name, activity_group.description))
            else:
                log.error('No activity groups defined - configure system correctly')
            raise Exception('ActivityGroup "%s" is not defined' % name)
        return activity_group

    def _delete_journals(self, s, activity_group, first_timestamp):
        # need to iterate because sqlite doesn't support multi-table delete and we have inheritance.
        for journal in s.query(ActivityJournal). \
                filter(ActivityJournal.activity_group == activity_group,
                       ActivityJournal.start == first_timestamp).all():
            Timestamp.clear(s, owner=self.owner_out, key=journal.id)
            s.delete(journal)
        s.flush()

    def _load_constants(self, s, ajournal):
        if self.constants:
            for constant in self.constants:
                name, value = constant.split('=', 1)
                log.debug(f'Setting {name}={value}')
                StatisticJournalText.add(log, s, name, None, None, self.owner_out, ajournal.activity_group,
                                         ajournal, value, ajournal.start)

    def _load_data(self, s, loader, data):

        ajournal, activity_group, first_timestamp, path, records = data
        timespan, warned, last_timestamp = None, 0, to_time(0.0)

        for record in records:
            if record.name == 'event' or (record.name == 'record' and record.timestamp > last_timestamp):
                if record.name == 'event' and record.value.event == 'timer' and record.value.event_type == 'start':
                    if timespan:
                        log.warning('Ignoring start with no corresponding stop (possible lost data?)')
                    else:
                        timespan = add(s, ActivityTimespan(activity_journal=ajournal,
                                                           start=record.value.timestamp,
                                                           finish=record.value.timestamp))
                if record.name == 'record':
                    lat, lon, timestamp = None, None, record.value.timestamp
                    # customizable loader
                    for field, name, units, type in self.record_to_db:
                        value = record.data.get(field, None)
                        if value is not None:
                            value = value[0][0]
                            loader.add(name, units, None, activity_group, ajournal, value, timestamp, type)
                            if name == LATITUDE:
                                lat = value
                            elif name == LONGITUDE:
                                lon = value
                    # values derived from lat/lon
                    if lat is not None and lon is not None:
                        x, y = Point.from_latitude_longitude(lat, lon).meters
                        loader.add(SPHERICAL_MERCATOR_X, M, None, activity_group, ajournal, x, timestamp,
                                   StatisticJournalFloat)
                        loader.add(SPHERICAL_MERCATOR_Y, M, None, activity_group, ajournal, y, timestamp,
                                   StatisticJournalFloat)
                        if self.add_elevation:
                            elevation = self.__oracle.elevation(lat, lon)
                            if elevation:
                                loader.add(RAW_ELEVATION, M, None, activity_group, ajournal, elevation, timestamp,
                                           StatisticJournalFloat)
                if record.name == 'event' and record.value.event == 'timer' \
                        and record.value.event_type == 'stop_all':
                    if timespan:
                        timespan.finish = record.value.timestamp
                        ajournal.finish = record.value.timestamp
                        timespan = None
                    else:
                        log.debug('Ignoring stop with no corresponding start (possible lost data?)')
                if record.name == 'record':
                    last_timestamp = record.timestamp
            else:
                if record.name == 'record':
                    log.warning('Ignoring duplicate record data for %s at %s - some data may be missing' %
                                (path, record.timestamp))




import datetime as dt

from bokeh.io import output_file
from bokeh.layouts import column
from bokeh.plotting import show

from ch2.data import *
from ch2.stoats.names import _log
from ch2.uranus.decorator import template


@template
def health(local_time):

    f'''
    # Health: {local_time.split()[0]}
    '''

    '''
    $contents
    '''

    '''
    ## Load Data
    
    Open a connection to the database and load the data we require.
    '''

    s = session('-v2')
    health = std_health_statistics(s)

    '''
    ## Health and Fitness
    '''

    output_file(filename='/dev/null')

    fitness, fatigue = like(FITNESS_D_ANY, health.columns), like(FATIGUE_D_ANY, health.columns)
    colours = ['black'] * len(fitness) + ['red'] * len(fatigue)
    alphas = [1.0] * len(fitness) + [0.5] * len(fatigue)
    ff = multi_line_plot(900, 300, TIME, fitness + fatigue, health, colours, alphas=alphas)
    log_ff = multi_line_plot(900, 100, TIME, [_log(name) for name in fitness + fatigue], health, colours,
                             alphas=alphas, x_range=ff.x_range, y_label='Log FF')
    atd = multi_dot_plot(900, 200, TIME, [ACTIVE_TIME_H, ACTIVE_DISTANCE_KM], health, ['black', 'grey'], alphas=[1, 0.5],
                         x_range=ff.x_range, rescale=True)
    shr = multi_plot(900, 200, TIME, [DAILY_STEPS, REST_HR], health, ['grey', 'red'], alphas=[1, 0.5],
                     x_range=ff.x_range, rescale=True, plotters=[bar_plotter(dt.timedelta(hours=20)), dot_plotter()])
    show(column(ff, log_ff, atd, shr))




import time as t
from itertools import zip_longest
from pprint import PrettyPrinter
from re import split

from ..stoats.names import M, KM, PC

PALETTE_RAINBOW = [
    ('plain', 'light gray', 'black'), ('plain-focus', 'white', 'black'),
    ('em', 'white', 'black'),
    ('label', 'dark gray', 'black'),
    ('selected', 'black', 'light gray'), ('selected-focus', 'black', 'white'),
    ('unimportant', 'dark blue', 'black'), ('unimportant-focus', 'light blue', 'black'),
    ('error', 'dark red', 'black'), ('error-focus', 'light red', 'black'),
    ('bar', 'dark gray', 'black'), ('bar-focus', 'dark gray', 'black'),
    ('rank-1', 'light red', 'black'), ('rank-2', 'yellow', 'black'), ('rank-3', 'dark green', 'black'),
    ('rank-4', 'light blue', 'black'), ('rank-5', 'light gray', 'black'),
    ('zone-6', 'black', 'white'), ('zone-5', 'black', 'light red'), ('zone-4', 'black', 'yellow'),
    ('zone-3', 'black', 'dark green'), ('zone-2', 'black', 'light blue'), ('zone-1', 'black', 'light gray'),
    ('quintile-1', 'light gray', 'black'), ('quintile-2', 'light blue', 'black'), ('quintile-3', 'dark green', 'black'),
    ('quintile-4', 'yellow', 'black'), ('quintile-5', 'light red', 'black'),
]


def sign(x):
    if x == 0:
        return 0
    elif x > 0:
        return 1
    else:
        return -1


def sigfig(value, n=2):
    scale = 1
    while value >= 10 ** n:
        value /= 10
        scale *= 10
    while value < 10 ** (n-1):
        value *= 10
        scale /= 10
    return int(0.5 + value) * scale


def em(text):
    return 'em', text


def error(text):
    return 'error', text


def label(text):
    return 'label', text


def force_iterable(data):
    try:
        iter(data)
        return data
    except TypeError:
        return [data]


def unique(elements, key=lambda x: x):
    known = set()
    for element in elements:
        value = key(element)
        if value not in known:
            known.add(value)
            yield element


def datetime_to_epoch(datetime):
    return t.mktime(datetime.timetuple())


def interleave(sep, iter):
    for started, value in enumerate(iter):
        if started:
            yield sep
        yield value


# https://docs.python.org/3/library/itertools.html#itertools-recipes
def grouper(iterable, n, fillvalue=None):
    "Collect data into fixed-length chunks or blocks"
    # grouper('ABCDEFG', 3, 'x') --> ABC DEF Gxx"
    args = [iter(iterable)] * n
    return zip_longest(*args, fillvalue=fillvalue)


PP = PrettyPrinter(indent=0, depth=1, width=80, compact=True)


def short_str(x):
    text = PP.pformat(x)
    lines = text.splitlines(False)
    if len(lines) > 1:
        return lines[0][:20] + '...' + lines[-1][-20:]
    else:
        return text


def format_metres(dist):
    if dist < 1000:
        return str(int(dist)) + M
    else:
        return f'{dist/1000:.1f}{KM}'


def format_percent(pc):
    if pc > 0.5:
        return f'{pc:.1f}{PC}'
    else:
        return f'{pc:.2f}{PC}'





from logging import getLogger, DEBUG, Formatter, INFO, StreamHandler, WARNING
from logging.handlers import RotatingFileHandler
from os.path import join
from sys import exc_info
from traceback import format_tb

from ..commands.args import COMMAND, LOGS, PROGNAME, VERBOSITY, LOG

log = getLogger(__name__)


def make_log(args, tui=False):

    if not getLogger('ch2').handlers:

        level_unset = args[VERBOSITY] is None
        level = 4 if level_unset else args[VERBOSITY][0]
        level = 10 * (6 - level)

        file_formatter = Formatter('%(levelname)-8s %(asctime)s: %(message)s')
        name = args[LOG] if LOG in args and args[LOG] else (
                (args[COMMAND] if COMMAND in args and args[COMMAND] else PROGNAME) + f'.{LOG}')
        path = join(args.dir(LOGS), name)
        file_handler = RotatingFileHandler(path, maxBytes=1e6, backupCount=10)
        file_handler.setLevel(DEBUG)
        file_handler.setFormatter(file_formatter)

        slog = getLogger('sqlalchemy')
        slog.setLevel(WARNING)
        slog.addHandler(file_handler)

        mlog = getLogger('matplotlib')
        mlog.setLevel(INFO)
        mlog.addHandler(file_handler)

        blog = getLogger('bokeh')
        blog.setLevel(DEBUG)
        blog.addHandler(file_handler)

        tlog = getLogger('tornado')
        tlog.setLevel(INFO)
        tlog.addHandler(file_handler)

        clog = getLogger('ch2')
        clog.setLevel(DEBUG)
        clog.addHandler(file_handler)

        if not tui or not level_unset:
            stderr_formatter = Formatter('%(levelname)8s: %(message)s')
            stderr_handler = StreamHandler()
            stderr_handler.setLevel(level)
            stderr_handler.setFormatter(stderr_formatter)
            # slog.addHandler(stderr_handler)
            # mlog.addHandler(stderr_handler)
            blog.addHandler(stderr_handler)
            tlog.addHandler(stderr_handler)
            clog.addHandler(stderr_handler)


def log_current_exception():
    t, e, tb = exc_info()
    try:
        log.debug(f'Exception: {e}')
    except:
        pass
    log.debug(f'Type: {t}')
    log.debug('Traceback:\n' + ''.join(format_tb(tb)))
